from funasr import AutoModel

def transcribe_audio(audio_path):
    """
    基于funasr的sensevoicesmall模型进行带时间戳的音频转录
    
    参数:
    audio_path (str): 输入音频文件的路径
    
    返回:
    list: 包含转录文本及其时间戳的列表，每个元素是一个字典，包含 'text', 'start', 'end' 字段
    """
    # 初始化模型，使用 sensevoicesmall 进行语音识别，开启时间戳功能
    model = AutoModel(model="iic/speech_paraformer-large-vad-punc_asr_nat-zh-cn-16k-common-vocab8404-pytorch")
    # model = AutoModel(model="iic/sensevoicesmall")
    
    # 调用模型进行流式识别
    result = model.generate(input=audio_path, timestamp=True)
    print(result)
    
    # 解析结果，提取带时间戳的信息
    timestamped_transcriptions = []
    if result:
        for item in result[0]['st']:
            timestamped_transcriptions.append({
                'text': item['text'],
                'start': item['start'],
                'end': item['end']
            })
    
    return timestamped_transcriptions


if __name__ == "__main__":
    audio_path = r'C:\Users\User\Desktop\Praasper\data\test_audio.wav'  # 替换为你的音频文件路径
    result = transcribe_audio(audio_path)
    # print(result)



import time
from scipy.signal import butter, filtfilt

# 记录程序开始执行的时间
START_TIME = time.time()


def show_elapsed_time():
    """显示从程序开始执行到现在的时间差"""
    elapsed = time.time() - START_TIME
    minutes = int(elapsed // 60)
    seconds = elapsed % 60
    milliseconds = int((elapsed % 1) * 1000)
    return f"{minutes:02d}:{int(seconds):02d}:{milliseconds:03d}"



def bandpass_filter(data, lowcut, highcut, fs, order=4):
    nyquist = 0.5 * fs
    low = lowcut / nyquist
    high = highcut / nyquist

    low = max(0.01, low)
    high = min(nyquist - 0.01, high)
    if low == 0:
        b, a = butter(order, high, btype='low', output="ba")
        filtered_data = filtfilt(b, a, data)
    else:
        try:
            b, a = butter(order, [low, high], btype='bandpass', output="ba")
            filtered_data = filtfilt(b, a, data)
        except ValueError:  # 如果设置的最高频率大于了可接受的范围
            b, a = butter(order, low, btype='high', output="ba")
            filtered_data = filtfilt(b, a, data)
    return filtered_data
    

if __name__ == '__main__':
    print(extract_cvt('我', 'zh'))


import os
import test_whisper
import test_whisper_timestamped
import test_whisperx


path = r"E:\Corpus\Emilia\emilia_ja_dataset_test"
fnames = list(set([f.split(".")[0] for f in os.listdir(path)]))
fnames.sort()




for fname in fnames:

    txt_name = f"{fname}.txt"
    wav_name = f"{fname}.wav"

    txt_path = os.path.join(path, txt_name)
    wav_path = os.path.join(path, wav_name)

    if not os.path.exists(txt_path) or not os.path.exists(wav_path):
        continue

    # Test:
    # test_whisperx.transcribe_audio(wav_path)
    test_whisper_timestamped.transcribe_audio(wav_path)





import whisperx
import torch
import os
from textgrid import TextGrid, IntervalTier
import soundfile as sf
# torch.backends.cuda.matmul.allow_tf32 = False
# torch.backends.cudnn.allow_tf32 = False

def transcribe_audio(audio_path, language=None):
    # 设置设备
    device = "cuda" if torch.cuda.is_available() else "cpu"
    batch_size = 16  # 根据你的显存调整批次大小
    compute_type = "float16" if torch.cuda.is_available() else "float32"

    # 加载ASR模型
    model = whisperx.load_model("large-v3-turbo", device, compute_type=compute_type)

    # 转录音频
    if language is None:
        result = model.transcribe(audio_path, batch_size=batch_size)
    else:
        result = model.transcribe(audio_path, batch_size=batch_size, language=language)

    # 加载对齐模型并对齐转录结果

    model_a, metadata = whisperx.load_align_model(language_code=result["language"], device=device)
    result = whisperx.align(result["segments"], model_a, metadata, audio_path, device)


    # 获取音频文件的最大时长
    with sf.SoundFile(audio_path) as f:
        max_duration = len(f) / f.samplerate

    # 打印转录结果
    # 初始化合并后的词段列表
    merged_segments = []
    if result.get("word_segments"):
        prev_segment = result["word_segments"][0]
        for segment in result["word_segments"][1:]:
            if segment["start"] >= max_duration:
                continue
            if prev_segment["end"] >= segment["start"]:  # 检查前后两个seg是否相邻
                # 合并相邻词段

                merged_segment = {
                    "start": prev_segment["start"],
                    "end": segment["end"],
                    "word": prev_segment["word"] + segment["word"]
                }
                prev_segment = merged_segment
            else:
                merged_segments.append(prev_segment)
                prev_segment = segment
        merged_segments.append(prev_segment)
    # print(merged_segments)

    # 获取音频文件所在目录的父目录
    parent_dir = os.path.dirname(os.path.dirname(audio_path))
    # 拼接 test_results 目录路径
    test_results_dir = os.path.join(parent_dir, 'test_results')
    # 确保 test_results 目录存在
    os.makedirs(test_results_dir, exist_ok=True)
    # 从音频文件名获取基础名称
    base_name = os.path.basename(os.path.splitext(audio_path)[0])
    # 拼接完整的 TextGrid 文件路径，保存到 test_results 目录
    textgrid_path = os.path.join(test_results_dir, base_name + '_whisperx.TextGrid')

    # 创建 TextGrid 对象
    tg = TextGrid()
    # 创建 IntervalTier 对象
    tier = IntervalTier(name='words', minTime=0, maxTime=max_duration)



    # 添加每个合并后的词段到 IntervalTier
    for segment in merged_segments:
        if segment["start"] >= max_duration:
            continue
        if segment["end"] > max_duration:
            segment["end"] = max_duration
        if not segment["word"].strip():
            continue
        tier.add(segment['start'], segment['end'], segment['word'])


    # 将 IntervalTier 添加到 TextGrid
    tg.append(tier)
    # 保存 TextGrid 文件
    tg.write(textgrid_path)
    return result

if __name__ == "__main__":
    audio_path = r'C:\Users\User\Desktop\Praasper\data\test_audio.wav'  # 替换为你的音频文件路径
    transcribe_audio(audio_path)


import gc
import math
import os
import copy
import numpy as np
from sklearn.cluster import DBSCAN

try:
    from .tool_auto import ReadSound
    from ..utils import show_elapsed_time, bandpass_filter
except ImportError:
    from praasper.VAD.tool_auto import ReadSound
    from praasper.utils import show_elapsed_time, bandpass_filter

# plat = os.name.lower()
# check if ffmpeg exists in the system path or the pydub package can find it
# os.environ["PATH"] += os.pathsep + resource_path(f".\\ffmpeg\\{plat}")
# print(resource_path(f".\\ffmpeg\\{plat}"))

def autoPraditorWithTimeRange(params, audio_obj, which_set, stime=0, etime=-1,verbose=False):
    params = copy.deepcopy(params)
    if etime == -1:
        ans_tps = autoPraditor(params, audio_obj, which_set, verbose=False)

    else:
        ans_tps = autoPraditor(params, audio_obj[stime*1000:etime*1000], which_set, verbose=False)
        ans_tps = [tp + stime for tp in ans_tps if 5 < tp <ans_tps[-1] - 5]
    
    ans_tps = sorted(ans_tps)
    return ans_tps


def autoPraditor(params, audio_obj, which_set, verbose=False):
    # 导入数据，并且遵循一定之格式
    for xset in params:
        for item in params[xset]:
            try:
                params[xset][item] = eval(params[xset][item])
            except:
                pass

    params = params[which_set]  # 选择是onset还是offset

    # 固定某些值
    params["ratio"] = 0.9
    params["win_size"] = int(50 / 44100 * audio_obj.frame_rate)
    params["ref_len"] = params["win_size"] * 5
    params["penalty"] = 10
    # params["numValid"] = 0.025 * audio_obj.frame_rate # 25 ms


    # print(params)
    _answer_frames = []

    _audio_obj = audio_obj

    _audio_samplerate = audio_obj.frame_rate

    _audio_arr_filtered = bandpass_filter(
        np.array(audio_obj.get_array_of_samples()),
        lowcut=params["cutoff0"],
        highcut=params["cutoff1"],
        fs=_audio_obj.frame_rate
    )

    # Audio已经准备好
    # 接下来就是第一步：DBSCAN聚类找噪声片段


    # 1.1. 降采样
    # 把一秒钟的音频分成n=40份

    _dsFactor = _audio_obj.frame_rate // 40
    # _audio_arr_ds = _audio_arr_filtered

    # 用取余容易得到[:-0]，返回一个空list
    # 用整除解决这个问题
    if len(_audio_arr_filtered) % _dsFactor != 0:
        _audio_arr_filtered = _audio_arr_filtered[:-(len(_audio_arr_filtered) % _dsFactor)]
    
    # warning or auto change?
    if which_set == "offset":
        _audio_arr_filtered = np.flip(_audio_arr_filtered)

    # max_frm_num = len(_audio_arr_filtered)
    _audio_arr_ds = _audio_arr_filtered.reshape((len(_audio_arr_filtered) // _dsFactor, _dsFactor))
    _audio_arr_ds = np.max(_audio_arr_ds, axis=1)  # 用max方法降采样

    # print(_audio_arr_ds[10:1000])

    _points_array = np.array([
        [_audio_arr_ds[i] for i in range(len(_audio_arr_ds) - 1)],
        [_audio_arr_ds[i + 1] for i in range(len(_audio_arr_ds) - 1)]
    ]).T




    _eps = params["eps_ratio"] * float(np.max(np.sort(_audio_arr_ds)[:int(.8 * len(_audio_arr_ds))]))  # 找到合适的radius，防止异常值


    del _audio_arr_ds
    gc.collect()

    _min_samples = math.ceil(0.3/_dsFactor * _audio_obj.frame_rate) #math.ceil(2 / (target_audio_samplerate/44100) / (interval*2/4281))
    try:
        _cluster = DBSCAN(eps=_eps, min_samples=_min_samples, metric="manhattan").fit(_points_array)
        # print(_cluster.labels_)

        # _cluster = DBSCAN_(eps=_eps, min_samples=_min_samples).fit(_points_array)
        # print(_cluster.labels_)
    except MemoryError:
        print("not enough memory")
        return []




    # To look for the label with which the coordinate is closet to the zero point
    # xy值加起来最小值 -> 最接近零点
    noise_label = 0
    for i in range(0, len(set(_cluster.labels_))-1):
        if np.min(np.sum(_points_array[_cluster.labels_ == i], axis=1)) < np.min(np.sum(_points_array[_cluster.labels_ == noise_label], axis=1)):
            noise_label = i
    _points_confirmed = _points_array[_cluster.labels_ == noise_label]
    # print(_labels, noise_label)

    # 把最小cluster以下的所有点都囊括进来
    _points_compensation = np.array(range(len(_points_array)))[np.sum(np.square(_points_array), axis=1) <= np.mean(np.sum(np.square(_points_confirmed), axis=1))]
    # print(_points_confirmed)

    _labels = _cluster.labels_

    for i in _points_compensation:
        _labels[int(i)] = noise_label

    _labels = [noise_label] * 3 + [i for i in _labels] + [noise_label] * 3  # 这句干啥用的？？？
    _indices_confirmed = [i-3 for i in range(len(_labels)) if _labels[i] == noise_label]  # or labels[i] == -1]

    # print(_indices_confirmed)


    # gather sampled area | target area

    _indices_completed = []
    for i in _indices_confirmed:
        for j in range(3):  # 原来是4，改成了3，只加上012
            if (i + j) not in _indices_completed:
                _indices_completed.append(i + j)

    # print(_indices_completed)

    _onsets = []
    _offsets = []
    for i in range(min(_indices_completed), max(_indices_completed) + 2):  # +2取不到 +1必定不存在
        if i in _indices_completed and (i - 1) not in _indices_completed:
            _onsets.append(i)
        elif i not in _indices_completed and (i - 1) in _indices_completed:
            _offsets.append(i - 1)

    # print(f" > Silence/Noise Onsets: {onsets}")  # 打印所有无声范围的onset
    # print(f" > Silence/Noise Offsets: {offsets}")  # 打印所有无声范围的offset

    # onset和offset的数量一定一样(? not really)
    # 把它俩成对组合后，筛除掉其中长度不够的（noise too short）


    while True:
        _bad_onoffsets = []
        for i in range(len(_offsets)-1):
            if (_onsets[i+1] - _offsets[i]) * _dsFactor /_audio_samplerate < .1:
                _bad_onoffsets.append(_onsets[i+1])
                _bad_onoffsets.append(_offsets[i])
        if len(_bad_onoffsets) == 0:
            break

        _onsets = [i for i in _onsets if i not in _bad_onoffsets]
        _offsets = [i for i in _offsets if i not in _bad_onoffsets]
    _onoffsets = [(_onsets[i], _offsets[i]) for i in range(len(_onsets))]

    # print(_onoffsets)

    # now offset means sound offset (not silence offset)
    # onset means sound onset (not silence onset)
    # with ProcessPoolExecutor(max_workers=threads) as executor:
    #     results = list(executor.map(process_items_with_params, *parameters, chunksize=threads))
    # print(np.cuda.get_device_id())
    # np.cuda.device = 1
    # print("---↘")
    total_num = len(_onoffsets)
    for i, (__offset, __onset) in enumerate(_onoffsets):
        if verbose:
            print(f"\r[{show_elapsed_time()}] ({os.path.basename(audio_obj.fpath)}) Drawing {which_set}(s) ({i+1}/{total_num}, {int((i+1)/total_num*100)}%)")

        # -------------------------------------------------
        # 强制跳过条件 Skip Condition

        if __onset <= 0 - 3:
            continue

        if __offset >= len(_cluster.labels_) + 3:
            continue

        # -----------------------------------------------

        __offset = 0 if __offset <= 0 else __offset
        # print(__offset * _dsFactor+1, __onset * _dsFactor, __offset * _dsFactor, __onset * _dsFactor-1)
        try:
            __candidate_y1_area = abs(np.array(
                _audio_arr_filtered[__offset * _dsFactor+1:__onset * _dsFactor] -
                _audio_arr_filtered[__offset * _dsFactor:__onset * _dsFactor-1]
            ))

        except ValueError:
            continue  # hit the bottom with no more frames

        __sample_startpoint = int(np.argmin(__candidate_y1_area) + __offset * _dsFactor)
        __sample_endpoint = __sample_startpoint - params["ref_len"]
        if __sample_endpoint < 0:
            __sample_endpoint = 0

        try:
            __candidate_y1_area = abs(np.array(
                _audio_arr_filtered[__sample_endpoint+1:__sample_startpoint] -
                _audio_arr_filtered[__sample_endpoint:__sample_startpoint - 1]
            ))
        except ValueError:
            __sample_startpoint = __sample_endpoint + params["ref_len"]
            __candidate_y1_area = abs(np.array(
                _audio_arr_filtered[__sample_endpoint+1:__sample_startpoint] -
                _audio_arr_filtered[__sample_endpoint:__sample_startpoint - 1]
            ))


        __candidate_y1_area = np.sort(__candidate_y1_area)[:int(len(__candidate_y1_area) * params["ratio"])]


        __y1_threshold = float(np.sum(__candidate_y1_area) / (__sample_startpoint - __sample_endpoint) * params["amp"])
        __ref_midpoint = int(__offset*_dsFactor + (__onset-__offset) * _dsFactor * 0.8)  # 3/4偏移量

        if i < len(_onoffsets) - 1:
            __ref_midpoint_next = int(_onoffsets[i+1][0]*_dsFactor + (_onoffsets[i+1][1]-_onoffsets[i+1][0]) * _dsFactor * 0.8)
        else:
            __ref_midpoint_next = len(_audio_arr_filtered)  # 设置为音频最后一帧的位置

        if i > 0:
            __ref_midpoint_prev = int(_onoffsets[i-1][0]*_dsFactor + (_onoffsets[i-1][1]-_onoffsets[i-1][0]) * _dsFactor * 0.8)
        else:
            __ref_midpoint_prev = 0  # 设置为音频第一帧的位置


        if __ref_midpoint < __sample_startpoint:
            __ref_midpoint = __sample_startpoint
        # print(np.argmin(candidate_y1_area), ref_midpoint)

        # ----------------- Processing onset area --------------

        # print(y1_threshold)

        __countValidPiece = 0
        __countBadPiece = 0
        __countDSTime = -1

        _final_answer = None
        while __ref_midpoint + __countDSTime < __ref_midpoint_next:  # 遍历从midpoint到next midpoint之间的每一帧
            __countDSTime += 1

            __left_boundary = __ref_midpoint + __countDSTime - params["win_size"]
            __right_boundary = __ref_midpoint + __countDSTime


            try:
                __raw_value = abs(_audio_arr_filtered[__left_boundary:__right_boundary] - _audio_arr_filtered[__left_boundary-1:__right_boundary-1])
            except ValueError:
                break
            __raw_value.sort()
            __raw_value = __raw_value[:int(len(__raw_value) * params["ratio"])]

            __y1_value = sum(__raw_value)/len(__raw_value)

            if __y1_value > __y1_threshold:

                __countValidPiece += 1
            else:
                __countBadPiece += 1 #params["penalty"]

            if __countValidPiece - __countBadPiece * params["penalty"]  <= 0:
                __countValidPiece = 0
                __countBadPiece = 0

            elif __countValidPiece - __countBadPiece >= params["numValid"]:
                _final_answer = __ref_midpoint + __countDSTime - __countValidPiece - __countBadPiece

                if which_set == "offset":
                    _final_answer = len(_audio_arr_filtered) - (_final_answer +  len(_audio_arr_filtered) % _dsFactor)
                _answer_frames.append(_final_answer)
                break
        
        
        # if _final_answer is None:
            # continue

    # 对 _answer_frames 进行从小到大排序
    # print(_answer_frames)

    if which_set == "offset":
        _answer_frames.reverse()
    # print(_answer_frames)

    _answer_frames = list(set(_answer_frames))

    return [frm/_audio_samplerate for frm in _answer_frames]







import whisper_timestamped
from textgrid import TextGrid, IntervalTier
import os
import soundfile as sf

def transcribe_audio(audio_path, language=None):
    """
    使用 whisper_timestamped 对音频文件进行转录
    
    参数:
    audio_path (str): 音频文件的路径
    
    返回:
    dict: 包含转录结果的字典
    """
    model = whisper_timestamped.load_model("large-v3-turbo")
    if language is None:
        result = whisper_timestamped.transcribe(model, audio_path)
    else:
        result = whisper_timestamped.transcribe(model, audio_path, language=language)

    # 获取音频文件的最大时长
    with sf.SoundFile(audio_path) as f:
        max_duration = len(f) / f.samplerate
        
    # 创建一个新的 TextGrid 对象
    tg = TextGrid()

    # 创建一个 IntervalTier 用于存储转录文本
    tier = IntervalTier('Transcription', 0, max_duration)

    # 合并
    merged_segments = []
    prev_segment = result["segments"][0]
    for segment in result["segments"][1:]:
        if segment["start"] >= max_duration:
            continue
        if prev_segment["end"] >= segment["start"]:  # 检查前后两个seg是否相邻
            # 合并相邻词段
            merged_segment = {
                "start": prev_segment["start"],
                "end": segment["end"],
                "text": prev_segment["text"] + segment["text"]
            }
            prev_segment = merged_segment
        else:
            merged_segments.append(prev_segment)
            prev_segment = segment
    merged_segments.append(prev_segment)




    for segment in merged_segments:
        if segment["start"] >= max_duration:
            continue
        if segment["end"] > max_duration:
            segment["end"] = max_duration
        if not segment["text"].strip():
            continue
        tier.add(segment['start'], segment['end'], segment['text'])
    # 将 IntervalTier 添加到 TextGrid 中
    tg.append(tier)

    # 保存 TextGrid 文件，文件路径与音频文件同目录，名称添加 _transcription 后缀
    base_name = os.path.splitext(audio_path)[0]
    textgrid_path = base_name + '_transcription.TextGrid'
    # 获取音频文件所在目录的父目录
    parent_dir = os.path.dirname(os.path.dirname(audio_path))
    # 拼接 test_results 目录路径
    test_results_dir = os.path.join(parent_dir, 'test_results')
    # 确保 test_results 目录存在
    os.makedirs(test_results_dir, exist_ok=True)
    # 从音频文件名获取基础名称
    base_name = os.path.basename(os.path.splitext(audio_path)[0])
    # 拼接完整的 TextGrid 文件路径，保存到 test_results 目录
    textgrid_path = os.path.join(test_results_dir, base_name + '_whisper_timestamped.TextGrid')
    tg.write(textgrid_path)
    # print(f"TextGrid 文件已保存至: {textgrid_path}")



    return result


if __name__ == "__main__":
    audio_path = r'C:\Users\User\Desktop\Praasper\data\test_audio.wav'  # 替换为你的音频文件路径
    result = transcribe_audio(audio_path)
    # print(result)



from funasr import AutoModel
from funasr.utils.postprocess_utils import lang_dict, emoji_dict, emo_set, event_set
try:
    from .utils import show_elapsed_time
except ImportError:
    from praasper.utils import show_elapsed_time


class SelectWord:
    def __init__(self, model: str="iic/SenseVoiceSmall", vad_model: str="fsmn-vad", device: str="auto", infer_mode: str="direct"):
        print(f"[{show_elapsed_time()}] Initializing ASR {model}")
        # 自动检测设备
        if device == "auto":
            import torch
            device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device = device
        self.infer_mode = infer_mode
        self.kwargs = {}
        
        if infer_mode == "direct":
            # 使用 FunASRNano 直接模式
            from funasr.models.fun_asr_nano.model import FunASRNano
            # from model import FunASRNano
            self.model, self.kwargs = FunASRNano.from_pretrained(
                model=model,
                disable_pbar=True,
                # checkpoint_callback=False,
                # trust_remote_code=False,
                disable_tqdm=True,
                device=self.device,
            )
            self.model.eval()


        else:
            # 默认使用 AutoModel
            self.model = AutoModel(
                model=model,
                vad_model=vad_model,
                # punc_model="ct-punc",
                vad_kwargs={"max_single_segment_time": 30000},
                device=self.device,
                # disable_update=True,
                # disable_pbar=True,
                # disable_log=True,
                # use_timestamp=True
                download_model=False,
                # trust_remote_code=True,
            )
        

    def transcribe(self, input_path, lang: str="zh"):

        if self.infer_mode == "direct":
            # 使用 FunASRNano 的 inference 方法
            res = self.model.inference(data_in=[input_path], **self.kwargs)
            # print(res)
            # text = res[0][0]["text"]
        else:
            # 使用 AutoModel 的 generate 方法
            res = self.model.generate(
                input=input_path,
                language=lang, 
                use_itn=False,
                #hotword="必须使用中文输出",
                # batch_size_s=60,
                # merge_length_s=15,

            )
            # text = res[0]["text"]
        # print(res)
        # exit()
        # text = rich_transcription_postprocess_text_only(text)
        # return text
        return res



def rich_transcription_postprocess_text_only(s):
    """
    修改版的rich_transcription_postprocess函数，只保留文字内容，去除所有表情符号和事件标记
    """
    # 替换所有语言标记为空
    for lang in lang_dict:
        s = s.replace(lang, "")
    
    # 移除所有特殊标记
    for special_tag in emoji_dict:
        s = s.replace(special_tag, "")
    
    # 移除所有表情符号和事件符号
    for emo in emo_set:
        s = s.replace(emo, "")
    
    for event in event_set:
        s = s.replace(event, "")
    
    # 清理多余的空格
    s = s.replace("The.", " ")
    s = " ".join(s.split())  # 移除多余的空格
    
    return s.strip()



if __name__ == "__main__":
    # input_path = r"tmp/15-1_5.wav" # f"{model.model_path}/example/en.mp3"
    input_path = r"input_single/15-1.wav"
    text = get_text_from_audio(input_path)
    print(text)


import os
import shutil
import numpy as np
import gc
import torch
from itertools import product
import copy
import random
import jellyfish
import concurrent.futures
from tqdm import tqdm

try:
    from .utils import *
    from .process import *
    from .select_word import *
    # from .post_process import *

except ImportError:
    from praasper.utils import *
    from praasper.process import *
    from praasper.select_word import *
    # from praasper.post_process import *

os.environ['HF_ENDPOINT'] = 'https://hf-mirror.com'  # 设置镜像源
os.environ["TRANSFORMERS_VERBOSITY"] = "error"
# os.environ["DISABLE_TQDM"] = "1"

# 清空资源
def clear_resources():
    # 清理 CUDA 缓存
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()
    # 强制垃圾回收
    gc.collect()

class init_model:

    def __init__(
        self,
        ASR: str="iic/SenseVoiceSmall",
        infer_mode: str = "funasr",
        # LLM: str="Qwen/Qwen2.5-1.5B-Instruct",
        device: str= "cpu"
    ):

        # 验证 infer_mode 参数值
        allowed_modes = ["direct", "funasr"]
        if infer_mode not in allowed_modes:
            raise ValueError(f"infer_mode must be one of {allowed_modes}, got {infer_mode}")
        
        # if infer_mode == "direct" and ASR != "FunAudioLLM/Fun-ASR-Nano-2512":
        #     self.tokenizer = init_tokenizer(ASR)
        
        self.ASR = ASR
        self.infer_mode = infer_mode
        # self.LLM = LLM
        self.device = device


        print(f"[{show_elapsed_time()}] Trying device ({self.device})...")
        # 检测硬件
        if self.device == "auto":
            if torch.cuda.is_available():
                self.device = "cuda"
                print(f"[{show_elapsed_time()}] CUDA detected, using GPU.")
            else:
                self.device = "cpu"
                print(f"[{show_elapsed_time()}] CUDA not available, using CPU.")
        else:
            print(f"[{show_elapsed_time()}] Using device: {self.device}")


        self.model = SelectWord(
            model=self.ASR,
            infer_mode=self.infer_mode,
            device=self.device
        )

        # init_LLM(self.LLM)


        # self.g2p = G2PModel()

        self.params = {'onset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}, 'offset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}}


    def annote(
        self,
        input_path: str,
        seg_dur=10.,
        min_pause=0.2,
        skip_existing: bool=False,
        verbose: bool=False,
    ):
        if os.path.isdir(input_path):
            fnames = [os.path.splitext(f)[0] for f in os.listdir(input_path) if f.endswith('.wav')]
            print(f"[{show_elapsed_time()}] {len(fnames)} valid audio files detected in {input_path}")

        else:
            fnames = [os.path.splitext(os.path.basename(input_path))[0]]
            input_path = os.path.dirname(input_path)
            print(f"[{show_elapsed_time()}] {fnames[0]} is detected in {input_path}")


        if not fnames:
            return

        for idx, fname in enumerate(fnames):
            wav_path = os.path.join(input_path, fname + ".wav")

            dir_name = os.path.dirname(os.path.dirname(wav_path))
            tmp_path = os.path.join(dir_name, "tmp")
            # 获取输入文件夹的名称
            input_folder_name = os.path.basename(input_path)
            # 在output目录下创建与输入文件夹同名的子目录
            output_path = os.path.join(dir_name, "output", input_folder_name)
            final_path = os.path.join(output_path, os.path.basename(wav_path).replace(".wav", ".TextGrid"))


            file_info = f"[{show_elapsed_time()}] {idx+1}/{len(fnames)} ({os.path.basename(wav_path)})"

            # 检查结果文件是否已存在，如果存在且skip_existing为True则跳过处理
            if skip_existing and os.path.exists(final_path):
                print(f"{file_info} Result exists, skipped", end="\r")
                continue

            try:
                # 尝试加载音频文件
                audio_obj = ReadSound(wav_path)
            except Exception as e:
                print(f"[{show_elapsed_time()}] Error loading audio file {wav_path}: {str(e)}")
                continue

            # 仅在音频加载成功后创建临时目录（很正确）
            if os.path.exists(tmp_path):
                shutil.rmtree(tmp_path)
                if verbose:
                    print(f"[{show_elapsed_time()}] Temporary directory {tmp_path} removed.")
            os.makedirs(tmp_path, exist_ok=False)

            os.makedirs(output_path, exist_ok=True)


            # auto search best params
            self.auto_vad(
                wav_path=wav_path,
                min_pause=min_pause,
                file_info=file_info,
            )

            final_tg = TextGrid()
            final_tg.tiers.append(IntervalTier(name="words", minTime=0., maxTime=audio_obj.duration_seconds))
            final_tg.tiers[0].strict = False

            segments = segment_audio(audio_obj, segment_duration=seg_dur, params=self.params, min_pause=min_pause, file_info=file_info)

            def process_segments(segment):
            # for start, end in segments:
                start, end = segment
                # count += 1

                # print(f"[{show_elapsed_time()}] Processing segment: {start/1000:.3f} - {end/1000:.3f} ({count})")
                audio_clip = audio_obj[start:end]
                clip_path = os.path.join(tmp_path, os.path.basename(wav_path).replace(".wav", f"_{start}_{end}.wav"))
                audio_clip.save(clip_path)

                audio_clip_result = self.model.transcribe(clip_path)
                audio_clip_single_words = audio_clip_result[0][0]["ctc_timestamps"]
                # print(audio_clip_single_words)


                # try:
                vad_tg = get_vad(clip_path, params=self.params, min_pause=min_pause, verbose=verbose)
                # except Exception as e:
                #     print(f"[{show_elapsed_time()}] ({os.path.basename(clip_path)}) VAD Error: {e}")
                #     return [None, None, None]
                
                intervals = vad_tg.tiers[0].intervals
                valid_intervals = [interval for interval in intervals if interval.mark not in ["", None]]# and interval.maxTime - interval.minTime > min_speech]
                # print(valid_intervals)

                # 整体转录模式
                for word in audio_clip_single_words:
                    s_w = word["start_time"]
                    e_w = word["end_time"]
                    t_w = word['token']

                    best_interval = None
                    max_overlap = -1
                    min_distance = float('inf')
                    closest_interval = None

                    for interval in valid_intervals:
                        s_i, e_i = interval.minTime, interval.maxTime
                        
                        # 计算重叠时间
                        overlap_start = max(s_w, s_i)
                        overlap_end = min(e_w, e_i)
                        overlap = max(0, overlap_end - overlap_start)
                        
                        # 如果有重叠且比当前最大重叠大
                        if overlap > max_overlap:
                            max_overlap = overlap
                            best_interval = interval
                        
                        # 计算距离（无重叠时使用）
                        if overlap == 0:
                            # 计算单词与间隔的距离
                            if e_w <= s_i:
                                # 单词在间隔前
                                distance = s_i - e_w
                            else:
                                # 单词在间隔后
                                distance = s_w - e_i
                            
                            if distance < min_distance:
                                min_distance = distance
                                closest_interval = interval
                    
                    # 选择最佳间隔
                    if best_interval is not None:
                        selected_interval = best_interval
                    else:
                        selected_interval = closest_interval
                    
                    # 将单词添加到选定的间隔
                    if selected_interval is not None:
                        if selected_interval.mark == "+":
                            selected_interval.mark = ""
                        
                        selected_interval.mark += f"{t_w}"
                        # selected_interval.words.append((t_w, s_w, e_w))
                
                timestamps = []
                for idx, valid_interval in enumerate(valid_intervals):
                    s, e = valid_interval.minTime, valid_interval.maxTime

                    # 获取间隔的文本（由前面的单词拼接而成）
                    text = valid_interval.mark

                    text = purify_text(text)
                    if not text or text == "+":
                        continue
                

                    s_point = s + start/1000
                    e_point = e + start/1000

                    if e_point >= audio_obj.duration_seconds:
                        e_point = audio_obj.duration_seconds
                    
                    timestamps.append([s_point, e_point, text])

                return timestamps

            with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:  # 使用线程池并发处理
                result = list(tqdm(executor.map(process_segments, segments), total=len(segments), desc=f"{file_info} Processing segments", leave=False))

            #############################
            # 因为在建立textgrid的时候使用了strict=False的mode，有可能存在某个tier是重复的
            # 需要检查并调整
            #############################
            results = result[0]
            for r in result[1:]:
                results.extend(r)
            for s_point, e_point, text in results:
                if s_point is None or e_point is None or text is None:
                    continue
                final_tg.tiers[0].add(s_point, e_point, text)

            # 检查并合并重叠的interval
            tier = final_tg.tiers[0]
            i = 1
            while i < len(tier.intervals):
                prev = tier.intervals[i - 1]
                curr = tier.intervals[i]
                # 如果当前interval与上一个interval重叠
                if curr.minTime < prev.maxTime:
                    # 合并：取最早开始和最晚结束
                    new_min = prev.minTime
                    new_max = max(prev.maxTime, curr.maxTime)
                    # 替换上一个interval
                    prev.maxTime = new_max
                    prev.mark += curr.mark
                    # 删除当前interval
                    del tier.intervals[i]
                    # 从头重新检查
                    i = 1
                else:
                    i += 1
 
            # ----------------------------
            final_tg.write(final_path)
                
        if os.path.exists(tmp_path):
            shutil.rmtree(tmp_path)
        # print(f"--------------- Processing completed ---------------")


    def auto_vad(self, wav_path, min_pause=0.2, verbose=False, file_info=""):
        """
        自动选取最优的VAD参数，根据随机选取的10秒音频。

        """


        # print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD processing started...")


        dir_name = os.path.dirname(os.path.dirname(wav_path))
        tmp_path = os.path.join(dir_name, "tmp")

        audio_obj = ReadSound(wav_path)
        if verbose:
            print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Full audio duration: {audio_obj.duration_seconds:.3f}")

        if audio_obj.duration_seconds > 10:
            # 计算最大起始时间
            max_start = audio_obj.duration_seconds - 10
            # 随机生成起始时间
            start_time = random.uniform(0, max_start)
            end_time = start_time + 10
        else:
            # 音频不足时间，选取整个音频
            start_time = 0
            end_time = audio_obj.duration_seconds
        
        # 截取选定的音频段
        selected_audio = audio_obj[start_time*1000:end_time*1000]
        if verbose:
            print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Selected audio: {start_time:.3f} - {end_time:.3f}")
        
        # 确保临时目录存在
        os.makedirs(tmp_path, exist_ok=True)
        # 将selected_audio保存到tmp文件夹
        base_name = os.path.splitext(os.path.basename(wav_path))[0]
        selected_audio_path = os.path.join(tmp_path, f"{base_name}_selected_{start_time:.0f}_{end_time:.0f}.wav")
        selected_audio.save(selected_audio_path)


        standard_result = self.model.transcribe(selected_audio_path)
        standard_transcript = standard_result[0][0]["text_tn"]  # text_tn 没有标点符号


        # params = default_params
        # params["offset"] = params["onset"]  # VAD模式特供


        def generate_param_grid(params):
            """生成参数组合列表"""
            keys = params.keys()
            values = params.values()
            for combination in product(*values):
                yield dict(zip(keys, combination))


        # df_res = pd.DataFrame(columns=["params", "transcript", "similarity"])
        # res = {}
        result = []

        # 使用示例
        param_grid = {
            'amp': np.arange(1.1, 1.5, 0.2),  #！找最多interval的amp
            "cutoff0": [0, 200],#range(0, 400, 200),
            'cutoff1': [min(audio_obj.frame_rate//2, 10800)],
            "numValid": [2000],#[int(min_speech/2*audio_obj.frame_rate//2)],

            'eps_ratio': np.arange(0.01, 0.15, 0.04)
        }
        def grid_search_optimal_params(params_replace):
        # for params_replace in generate_param_grid(param_grid):
            # res_key = (params_replace["amp"], params_replace["cutoff0"])
            adjusted_params = default_params.copy()
            adjusted_params["offset"] = adjusted_params["onset"]
            # print(adjusted_params)
            for p in params_replace:
                adjusted_params["onset"][p] = str(params_replace[p])
            # print(params_replace)
            adjusted_params["offset"] = adjusted_params["onset"]

            # print(adjusted_params)
            # print(f"[{show_elapsed_time()}] Testing {adjusted_params["onset"]}")
            # exit()
            
            onsets = autoPraditorWithTimeRange(adjusted_params, selected_audio, "onset", verbose=False)
            offsets = autoPraditorWithTimeRange(adjusted_params, selected_audio, "offset", verbose=False)

            onsets = sorted(onsets)
            offsets = sorted(offsets)


            if verbose:   
                print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD onsets: {onsets}")
                print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD offsets: {offsets}")
                


            if onsets or offsets:
                if not onsets:
                    onsets = [0.0]
                if not offsets:
                    offsets = [selected_audio.duration_seconds]


                if onsets[0] >= offsets[0]:
                    onsets = [0.0] + onsets
                
                if offsets[-1] <= onsets[-1]:
                    offsets.append(selected_audio.duration_seconds)

                # Select the one offset that is closest to onset and earlier than onset
                valid_onsets = []
                valid_offsets = []

                if len(onsets) <= len(offsets):
                    for i, onset in enumerate(onsets):
                        # print(onset)
                        if i == 0:
                            valid_offsets.append(offsets[-1])  # 最后一个offset
                            valid_onsets.append(onsets[0])  # 第一个offset
                        else:
                            try:
                                valid_offsets.append(max([offset for offset in offsets if onsets[i-1] < offset < onset]))  # 不会影响到最后一个offset，应为需要夹在onset中间
                                valid_onsets.append(onset)  # 下一个onset

                            except ValueError:
                                pass
                else:  # len(onsets) > len(offsets)
                    reversed_offsets = list(reversed(offsets))
                    for i, reversed_offset in enumerate(reversed_offsets):
                        if i == 0:
                            valid_onsets.append(onsets[0])
                            valid_offsets.append(reversed_offset)
                        else:
                            try:
                                valid_onsets.append(min([onset for onset in onsets if reversed_offset < onset < reversed_offsets[i-1]]))
                                valid_offsets.append(reversed_offset)

                            except ValueError:
                                pass
                    valid_offsets = list(reversed(valid_offsets))
                


                onsets = sorted(valid_onsets)
                offsets = sorted(valid_offsets)
                
                num_intervals = len(onsets)

                # 根据min pause 调整onset和offset
                bad_onsets = []
                bad_offsets = []

                for i in range(len(onsets)-1):
                    if onsets[i+1] - offsets[i] < min_pause:
                        bad_onsets.append(onsets[i+1])
                        bad_offsets.append(offsets[i])

                onsets = [x for x in onsets if x not in bad_onsets]
                offsets = [x for x in offsets if x not in bad_offsets]

                os.makedirs(tmp_path, exist_ok=True)  # 确保临时目录存在

                audio_obj_clipped = None
                for i, (onset, offset) in enumerate(zip(onsets, offsets)):
                    if audio_obj_clipped is None:
                        audio_obj_clipped = selected_audio[onset*1000:offset*1000]
                    else:
                        audio_obj_clipped += selected_audio[onset*1000:offset*1000]
                    
                    if i != len(onsets)-1:
                        # 创建500ms空白音频
                        silence_duration = 0.2  # 500ms
                        silence_samples = int(silence_duration * selected_audio.frame_rate)
                        silence_arr = np.zeros(silence_samples, dtype=np.float32)
                        silence_audio = ReadSound(fpath=None, arr=silence_arr, duration_seconds=silence_duration, frame_rate=selected_audio.frame_rate)
                        audio_obj_clipped += silence_audio 
                # 为裁剪的音频文件生成唯一的文件名
                base_name = os.path.splitext(os.path.basename(wav_path))[0]
                clip_path = os.path.join(tmp_path, f"{base_name}_clip_{i}.wav")
                audio_obj_clipped.save(clip_path)

            
                # 转录并获取结果
                clip_result = self.model.transcribe(clip_path)
                clip_transcript = clip_result[0][0]["text_tn"]

                similarity = jellyfish.jaro_winkler_similarity(clip_transcript, standard_transcript)
                similarity = min(similarity, 0.9)
                
            else:
                similarity = 0.
                num_intervals = 0  # 可能有潜在bug，即所有num_intervals都等于0（日后待修）

            # 复制两个变量
            num_intervals_copy = num_intervals
            adjusted_params_copy = copy.deepcopy(adjusted_params)

            # 然后可以像原来一样使用复制后的变量
            return [num_intervals_copy, similarity, adjusted_params_copy]
            # result.append([num_intervals_copy, similarity, adjusted_params_copy])
        
        with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:  # 使用线程池并发处理
            result = list(tqdm(executor.map(grid_search_optimal_params, generate_param_grid(param_grid)),
                              total=len(list(generate_param_grid(param_grid))),
                              desc=f"{file_info} Locate optimal params", leave=False))
        
        max_result = max(result, key=lambda x: (x[1], x[0]))
        max_item, best_params = max_result[0], max_result[2]
        if verbose:
            print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD best numIntervals: {max_item}, params: {best_params['onset']}")

        self.params = best_params

    def release_resources(self):
        """释放模型资源"""
        if hasattr(self, 'model') and self.model is not None:
            # 如果是直接模式，尝试释放模型
            if hasattr(self.model, 'to'):
                self.model.to('cpu')
            # 清空模型引用
            del self.model
        # 清空资源
        clear_resources()
        print("Resources released successfully")

# 导出clear_resources函数
__all__ = ['init_model', 'clear_resources']


from textgrid import IntervalTier, TextGrid

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

try:
    from utils import *
    from VAD.core_auto import *

except ImportError:
    from praasper.utils import *
    from praasper.VAD.core_auto import *
import os

import unicodedata
from tqdm import tqdm

default_params = {'onset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}, 'offset': {'amp': '1.47', 'cutoff0': '60', 'cutoff1': '10800', 'numValid': '475', 'eps_ratio': '0.093'}}



def purify_text(text):
    """
    清理文本中的无效字符，保留所有语言的文字字符
    
    :param text: 输入的文本
    :return: 清理后的文本
    """

    text = text.strip()
    # 只删除标点符号，保留所有语言的文字字符
    text = ''.join('' if unicodedata.category(c).startswith('P') else c for c in text)
    return text


def segment_audio(audio_obj, segment_duration=10, min_pause=0.2, params="folder", verbose=False, file_info=""):
    wav_path = audio_obj.fpath
    if verbose:
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) Start segmentation (<= {segment_duration}s)...")


    audio_obj = ReadSound(wav_path)

    folder_param_path = os.path.join(os.path.dirname(wav_path), "params_vad.txt")
    file_txt_path = wav_path.replace(".wav", "_vad.txt")

    match params:
        case "file":
            if os.path.exists(file_txt_path):
                with open(file_txt_path, "r") as f:
                    params = eval(f.read())
            else:
                params = default_params
        case "folder":
            if os.path.exists(folder_param_path):
                with open(folder_param_path, "r") as f:
                    params = eval(f.read())
            else:
                params = default_params
        case "default":
            params = default_params
        case _:  # 最好直接输入dict
            if type(params) == dict:
                params = params
            elif type(params) == str:
                params = eval(params)
            else:
                params = default_params
    
    params["offset"] = params["onset"]  # VAD特供

    segments = []

    y = audio_obj.arr
    sr = audio_obj.frame_rate

    audio_len = len(y) /sr

    start = 0.0 * 1000
    end = segment_duration * 1000
    total_length = audio_len * 1000
    
    desc = f"{file_info} Segmenting..."
    with tqdm(total=total_length, desc=desc, unit="%", bar_format="{l_bar}{bar}", leave=False) as pbar:
        while end <= total_length:
            segment = audio_obj[start:end]
            # print(type(segment) == type(audio_obj))
            onsets = autoPraditorWithTimeRange(params, segment, "onset", verbose=False)
            offsets = autoPraditorWithTimeRange(params, segment, "offset", verbose=False)
            # print()
            # print(start, end)
            # print(onsets, offsets, total_length)
            if not onsets or not offsets:
                try:
                    segments[-1][1] = end
                except IndexError:
                    pass
                # continue
            else:

                ################
                # 找到一个尽可能大的pause，使得onset和offset之间的差值大于pause
                ################

                # 从最后一个onset开始往前遍历
                tmp_pause = min_pause
                found = False
                while tmp_pause > 0 and not found:
                    for i in range(len(onsets)-1, 0, -1):
                        current_onset = onsets[i]

                        # 找到当前onset之前的最后一个offset
                        try:
                            prev_offset = [offset for offset in offsets if offset < current_onset][-1]
                        except IndexError:
                            continue

                        if current_onset - prev_offset > tmp_pause:
                            # 若差值大于tmp_pause，则取得他们的均值
                            target_offset = (current_onset + prev_offset) / 2
                            end = start + target_offset * 1000
                            found = True
                            break
                    if not found:
                        tmp_pause -= 0.01
                # -----------------

                segments.append([start, end])

            start = end
            end = start + segment_duration * 1000
            pbar.n = start
            pbar.refresh()

            if end > total_length and end > start:
                if total_length - start < 10:
                    segments[-1][1] = total_length
                    pbar.n = total_length
                    pbar.refresh()
                    break
                else:
                    segments.append([start, total_length])
                    pbar.n = total_length
                    pbar.refresh()
                    break
    # print(segments)
    if not segments:
        segments.append([0.0, total_length])
    
    # print(segments)
    # exit()
    return segments


def get_vad(wav_path, min_pause=0.2, params="folder", if_save=False, verbose=False):
    # print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD processing started...")


    audio_obj = ReadSound(wav_path)

    folder_param_path = os.path.join(os.path.dirname(wav_path), "params_vad.txt")
    file_txt_path = wav_path.replace(".wav", "_vad.txt")

    match params:
        case "file":
            if os.path.exists(file_txt_path):
                with open(file_txt_path, "r") as f:
                    params = eval(f.read())
            else:
                params = default_params
        case "folder":
            if os.path.exists(folder_param_path):
                with open(folder_param_path, "r") as f:
                    params = eval(f.read())
            else:
                params = default_params
        case "default":
            params = default_params
        case _:  # 最好直接输入dict
            if type(params) == dict:
                params = params
            elif type(params) == str:
                params = eval(params)
            else:
                params = default_params
    
    params["offset"] = params["onset"]  # VAD模式特供

    # print(params)


    onsets = autoPraditorWithTimeRange(params, audio_obj, "onset", verbose=False)
    offsets = autoPraditorWithTimeRange(params, audio_obj, "offset", verbose=False)

    onsets = sorted(onsets)
    offsets = sorted(offsets)


    if verbose:   
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD onsets: {onsets}")
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD offsets: {offsets}")
        
    tg = TextGrid()
    interval_tier = IntervalTier(name="interval", minTime=0., maxTime=audio_obj.duration_seconds)
    tg.append(interval_tier)

    if onsets or offsets:
        if not onsets:
            onsets = [0.0]
        if not offsets:
            offsets = [audio_obj.duration_seconds]
    else:
        return tg
    
    if onsets[0] >= offsets[0]:
        onsets = [0.0] + onsets
    
    if offsets[-1] <= onsets[-1]:
        offsets.append(audio_obj.duration_seconds)

    # Select the one offset that is closest to onset and earlier than onset
    valid_onsets = []
    valid_offsets = []

    if len(onsets) <= len(offsets):
        for i, onset in enumerate(onsets):
            # print(onset)
            if i == 0:
                valid_offsets.append(offsets[-1])  # 最后一个offset
                valid_onsets.append(onsets[0])  # 第一个offset
            else:
                try:
                    valid_offsets.append(max([offset for offset in offsets if onsets[i-1] < offset < onset]))  # 不会影响到最后一个offset，应为需要夹在onset中间
                    valid_onsets.append(onset)  # 下一个onset

                except ValueError:
                    pass
    else:  # len(onsets) > len(offsets)
        reversed_offsets = list(reversed(offsets))
        for i, reversed_offset in enumerate(reversed_offsets):
            if i == 0:
                valid_onsets.append(onsets[0])
                valid_offsets.append(reversed_offset)
            else:
                try:
                    valid_onsets.append(min([onset for onset in onsets if reversed_offset < onset < reversed_offsets[i-1]]))
                    valid_offsets.append(reversed_offset)

                except ValueError:
                    pass
        valid_offsets = list(reversed(valid_offsets))
    


    onsets = sorted(valid_onsets)
    offsets = sorted(valid_offsets)



    bad_onsets = []
    bad_offsets = []

    for i in range(len(onsets)-1):
        if onsets[i+1] - offsets[i] < min_pause:
            bad_onsets.append(onsets[i+1])
            bad_offsets.append(offsets[i])

    onsets = [x for x in onsets if x not in bad_onsets]
    offsets = [x for x in offsets if x not in bad_offsets]
    
    for onset, offset in zip(onsets, offsets):
        interval_tier.add(onset, offset, "+")


    tg.append(interval_tier)
    tg.write(wav_path.replace(".wav", "_vad.TextGrid"))  # 将TextGrid对象写入文件

    tg = TextGrid()
    tg.read(wav_path.replace(".wav", "_vad.TextGrid"))

    if not if_save:
        os.remove(wav_path.replace(".wav", "_vad.TextGrid"))
    else:
        print(f"[{show_elapsed_time()}] ({os.path.basename(wav_path)}) VAD results saved")
    
    
    return tg


if __name__ == "__main__":
    # segments = segment_audio("data/test_audio.wav", segment_duration=3.5)
    # print(segments)
    auto_vad("data/test_audio.wav", min_pause=0.2, if_save=True, verbose=True)


import numpy as np
import librosa


class ReadSound:
    def __init__(self, fpath=None, arr=None, duration_seconds=None, frame_rate=None):

        self.fpath = fpath
        if fpath is None:
            if arr is None:
                raise ValueError("Need audio input. Receive None.")
            else:
                self.arr = arr
                self.duration_seconds = duration_seconds
                self.frame_rate = frame_rate

        else:  # 如果有fpath
            if arr is not None:
                self.arr = arr
                self.duration_seconds = duration_seconds
                self.frame_rate = frame_rate
            else:
                try:
                    self.arr, self.frame_rate = librosa.load(fpath, sr=None, dtype=np.float32)
                    # self.arr = (self.arr * 32767).astype('int16')  # 转换为 int16 类型
                    self.duration_seconds = librosa.get_duration(y=self.arr, sr=self.frame_rate)
                except Exception as e:
                    raise IOError(f"Failed to load audio file {fpath}: {str(e)}") from e



        try:
            self.arr = self.arr[:, 0]
        except IndexError:
            pass

        self.max = np.max(np.abs(self.arr))

    def __getitem__(self, ms):


        start = int(ms.start * self.frame_rate / 1000) if ms.start is not None else 0
        end = int(ms.stop * self.frame_rate / 1000) if ms.stop is not None else len(self.arr)

        start = min(start, len(self.arr))
        end = min(end, len(self.arr))

        return ReadSound(fpath=self.fpath, arr=self.arr[start:end], duration_seconds=(end - start) / self.frame_rate, frame_rate=self.frame_rate)

    def power(self):
        """
        计算整段信号的平均功率
        
        :return: 整段信号的平均功率
        """
        return np.mean(self.arr ** 2)
    
    def min_power_segment(self, segment_duration=1.0):
        """
        计算整段信号中每个segment_duration时长的最小功率
        
        :param segment_duration: 每个segment的时长，单位为秒
        :return: 每个segment的最小功率数组
        """
        if segment_duration > self.duration_seconds:
            # print("Segment duration must be shorter than audio duration.")
            segment_duration = self.duration_seconds
        num_segments = int(self.duration_seconds // segment_duration)
        segment_powers = []
        step = max(1, num_segments // 2)  # step 为窗口一半，最小为1
        for i in range(0, num_segments, step):
            start = int(i * segment_duration * self.frame_rate)
            end = int((i + 1) * segment_duration * self.frame_rate)
            segment = self.arr[start:end]
            # ReadSound(fpath=self.fpath, arr=self.arr[start:end], duration_seconds=(end - start) / self.frame_rate, frame_rate=self.frame_rate)
            segment_powers.append(np.min(segment ** 2))
            timestamps = [[start, end]]
        # 找到最小功率的索引
        min_power_idx = np.argmin(segment_powers)
        start, end = timestamps[min_power_idx]
        # 返回最小功率段的 ReadSound 对象
        return ReadSound(
            fpath=self.fpath,
            arr=self.arr[start:end],
            duration_seconds=(end - start) / self.frame_rate,
            frame_rate=self.frame_rate
        )


    def get_array_of_samples(self):
        return self.arr
    
    def save(self, fpath):
        """
        使用 soundfile 保存音频文件

        :param fpath: 保存音频文件的路径
        """
        import soundfile as sf
        sf.write(fpath, self.arr, self.frame_rate)

    def __add__(self, other):
        """
        将两个ReadSound对象相加
        
        :param other: 另一个ReadSound对象
        :return: 新的ReadSound对象，包含相加后的音频数据
        """
        if not isinstance(other, ReadSound):
            raise TypeError("Only ReadSound objects can be added together.")
        
        # 检查采样率是否相同
        if self.frame_rate != other.frame_rate:
            # 如果采样率不同，将other重采样到self的采样率
            other_resampled = librosa.resample(
                other.arr, 
                orig_sr=other.frame_rate, 
                target_sr=self.frame_rate
            )
            other_arr = other_resampled
        else:
            other_arr = other.arr
        
        # 将两个音频数组相加
        combined_arr = np.concatenate([self.arr, other_arr])
        
        # 计算新的时长
        combined_duration = self.duration_seconds + other.duration_seconds
        
        # 创建新的ReadSound对象
        return ReadSound(
            fpath=None,
            arr=combined_arr,
            duration_seconds=combined_duration,
            frame_rate=self.frame_rate
        )

    def reverse(self):
        """
        反转音频数据的时间顺序
        
        :return: 新的ReadSound对象，包含反转后的音频数据
        """
        # 使用numpy的flip函数反转数组
        reversed_arr = np.flip(self.arr)
        
        # 创建新的ReadSound对象
        return ReadSound(
            fpath=None,
            arr=reversed_arr,
            duration_seconds=self.duration_seconds,
            frame_rate=self.frame_rate
        )


def isAudioFile(fpath):
    # 所有的音频后缀
    audio_extensions = [
        '.mp3',  # MPEG Audio Layer-3
        '.wav',   # Waveform Audio File Format
        ".WAV",
        '.ogg',   # Ogg
        '.flac',  # Free Lossless Audio Codec
        '.aac',   # Advanced Audio Codec
        '.m4a',   # MPEG-4 Audio Layer
        '.alac',  # Apple Lossless Audio Codec
        '.aiff',  # Audio Interchange File Format
        '.au',    # Sun/NeXT Audio File Format
        '.aup',   # Audio Unix/NeXT
        '.ra',    # RealAudio
        '.ram',   # RealAudio Metafile
        '.rv64',  # Raw 64-bit float (AIFF/AIFF-C)
        '.spx',   # Ogg Speex
        '.voc',   # Creative Voice
        '.webm',  # WebM (audio part)
        '.wma',   # Windows Media Audio
        '.xm',    # FastTracker 2 audio module
        '.it',    # Impulse Tracker audio module
        '.mod',   # Amiga module (MOD)
        '.s3m',   # Scream Tracker 3 audio module
        '.mtm',   # MultiTracker audio module
        '.umx',   # FastTracker 2 extended module
        '.dxm',   # Digital Tracker (DTMF) audio module
        '.f4a',   # FAudio (FMOD audio format)
        '.opus',  # Opus Interactive Audio Codec
    ]
    if any(fpath.endswith(ext) for ext in audio_extensions):
        return True
    else:
        return False




import whisper
from textgrid import TextGrid, IntervalTier
import os
from itertools import product
import torch
import gc
import soundfile as sf

def transcribe_audio(audio_path, language=None, model_name='large-v3-turbo', word_timestamps=False):
    """
    使用Whisper模型转录指定音频文件，并将结果输出为TextGrid格式。

    参数:
    audio_path (str): 音频文件的路径。
    model_name (str): Whisper模型的名称，默认为 'large-v3-turbo'。
    word_timestamps (bool): 是否包含单词级时间戳，默认为True。

    返回:
    TextGrid: 包含转录结果的TextGrid对象。
    """
    device = 'cuda'
    model = whisper.load_model(model_name, device=device)

    # 获取音频文件的最大时长
    with sf.SoundFile(audio_path) as f:
        max_duration = len(f) / f.samplerate


    # 转录音频
    if language is None:
        result = model.transcribe(audio_path, word_timestamps=word_timestamps)
    else:
        result = model.transcribe(audio_path, word_timestamps=word_timestamps, language=language)

    # 创建一个空列表来存储转录结果
    transcription_list = []
    
    # 将每个分段添加到层级中
    if not result['segments']:
        return transcription_list
        
    target_seg_tiername = "word_segments" if "word_segments" in result else "segments"

    textname = "word" if "word" in result[target_seg_tiername][0] else "text"
    
    merged_segments = []
    prev_segment = result[target_seg_tiername][0]
    for segment in result[target_seg_tiername][1:]:
        if segment["start"] >= max_duration:
            continue
        if prev_segment['end'] >= segment['start']:  # 检查是否相邻
            prev_segment = {
                'start': prev_segment['start'],
                'end': segment['end'],
                'text': prev_segment[textname] + segment[textname]
            }
        else:
            merged_segments.append(prev_segment)
            prev_segment = segment
    merged_segments.append(prev_segment)

    # 创建 TextGrid 对象
    tg = TextGrid()
    # 创建 IntervalTier 对象用于存储转录结果
    tier = IntervalTier(name='transcription', minTime=0, maxTime=max_duration)

    for segment in merged_segments:
        if segment["start"] >= max_duration:
            continue
        if segment["end"] > max_duration:
            segment["end"] = max_duration
        if not segment[textname].strip():
            continue
        # print(segment['start'], segment['end'], segment[textname])
        tier.add(segment['start'], segment['end'], segment[textname])
    tg.append(tier)

    # 保存 TextGrid 文件，保存到 test_results 目录
    # 获取音频文件所在目录的父目录
    parent_dir = os.path.dirname(os.path.dirname(audio_path))
    # 拼接 test_results 目录路径
    test_results_dir = os.path.join(parent_dir, 'test_results')
    # 确保 test_results 目录存在
    os.makedirs(test_results_dir, exist_ok=True)
    # 从音频文件名获取基础名称
    base_name = os.path.basename(os.path.splitext(audio_path)[0])
    # 拼接完整的 TextGrid 文件路径，保存到 test_results 目录
    textgrid_path = os.path.join(test_results_dir, base_name + f'_whisper_word_timestamp_{word_timestamps}.TextGrid')
    tg.write(textgrid_path)
    
    return transcription_list, result


if __name__ == '__main__':
    audio_path = r"E:\Corpus\ma\audio\32-4.wav"
    word_timestamps = [True, False]

    results = {}

    for word_timestamps in [False, True]:
        if word_timestamps:
            mode = 'with_timestamps'
        else:
            mode = 'no_timestamps'

        textgrid, result = transcribe_audio(audio_path, language="zh", word_timestamps=word_timestamps)

        results['whisper'] = results.get('whisper', {})
        results['whisper'][mode] = [(segment['start'], segment['end'], segment['text']) for segment in result['segments']]

        print(results['whisper'][mode])




from setuptools import setup, find_packages
import os

# 读取README.md文件作为长描述
with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()

# 读取requirements.txt文件作为依赖
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), 'r', encoding='utf-8') as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name="praasper",
    version='0.5.5',
    author="Tony Liu",
    author_email="paradeluxe3726@gmail.com",
    description="A tool for automatic speech recognition and annotation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/praasper/praasper",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=requirements,

)